/* cache.c   simulate cache design

   P. Chongstitvatana	8 Sept 1997
   update				8 Feb 2012
*/

#include <stdio.h>
#include <stdlib.h>

#define MAX 	1000	// max number of slot
#define MAXWAY  4

int  W,CS,M,H,LRU;		// W linesize, CS cache size, M way, H num of slot
int  tag[MAXWAY][MAX];
char lru[MAXWAY][MAX];
FILE *fn;
int hit, total, clock;

int linesize[] = {1,2,4,8};
int way[] = {1,2,4};
int cachesize[] = {128,256,512,1024};

double ratio;

// search all sets for minimum lru (age)
int leastrecent(int idx){
	int i, k, min;
	min = lru[0][idx];
	k = 0;
	for( i=1; i<M; i++ )		// search all sets
		if( lru[i][idx] < min ){
			min = lru[i][idx];
			k = i;
		}
	return k;
}

// stamp lru with clock (age)
void updateLRU( int n, int idx ){
	lru[n][idx] = clock;
}

void replace( int idx, int tn ){
	int m;
	if ( LRU )
		m = leastrecent(idx);
//	else
//		m = lrand48() % M;
	tag[m][idx] = tn;
	updateLRU(m,idx);
}

int readtrace(FILE *fn){
	int a;
	if( fscanf(fn,"%d\n",&a) == EOF ) return EOF;
	return a;
}

// initialise tag and lru (age)
void init(void){
	int i,j;
	clock = 0;
	hit = 0;
	total = 0;
	H = CS/(M*W);				// number of slot
	for( j=0; j<M; j++ )
		for( i=0; i<H; i++ ) {
			tag[j][i] = 0;
			lru[j][i] = 0;
		}
}

void trace(void){
	int a,idx,tn,i;
	init();
	fn = fopen("trace.tr","rt");
	while( (a = readtrace(fn)) != EOF ) {
		total++;
		clock++;				// age, use to update LRU

		a = a / W;
		idx = a % H;			// point to slot
		tn = a / H;				// address tag
		for( i=0; i<M; i++ )	// search all sets (M sets)
			if( tag[i][idx] == tn ) {  // hit
				hit++;
				updateLRU(i,idx);
				i = M + 1;				// exit for
			}
		if( i == M ) replace(idx,tn); // miss
	}
	fclose(fn);
}

int main(void){
	int i,j,k,s;
	LRU = 1;
	for( k=0; k<4; k++) {
		CS = cachesize[k];
		for( j=0; j<3; j++) {
			M = way[j];
			printf("Cache size %d way %d\n",CS,M);
			for( i=0; i<4; i++ ) {
				W = linesize[i];
				trace();
				ratio = (double)hit / (double) total;
				printf("linesize %d slot %d hit ratio %5.3f\n",W,H,ratio);
			}
		}
	}
	return 0;
}
